from __future__ import annotations

from collections.abc import AsyncIterator, Sequence
from logging import getLogger
from ssl import SSLContext, PROTOCOL_TLS, CERT_REQUIRED
from platform import system
from types import TracebackType
from typing import Type, TypeVar

from aiohttp import (  # pylint: disable=unused-import
    ClientSession,
    ClientTimeout,
    TCPConnector,
    WSMsgType,
    WSMessage
)
from certifi import where as certifi_where


Self = TypeVar('Self', bound='WSUtils')


class WSUtils:
    __slots__ = (
        'logger',
        'websocket_timeout',
        'websocket_session',
        'connector'
    )

    def __init__(self) -> None:
        self.logger = getLogger(self.__class__.__name__)
        self.websocket_timeout = ClientTimeout(
            total=None, sock_connect=None, sock_read=None
        )
        self.websocket_session: ClientSession | None = None
        self.connector = TCPConnector(limit_per_host=5)

    @property
    def ssl_context(self) -> SSLContext:
        ssl_context = SSLContext(PROTOCOL_TLS)
        ssl_context.verify_mode = CERT_REQUIRED
        ssl_context.check_hostname = True
        ssl_context.load_default_certs()

        if system().lower() == 'darwin':
            ssl_context.load_verify_locations(cafile=certifi_where())
        return ssl_context

    async def __aenter__(self) -> Self:
        if not self.websocket_session or self.websocket_session.closed:
            self.logger.debug('Created a new session')
            self.websocket_session = ClientSession(
                timeout=self.websocket_timeout, connector=self.connector
            )
        return self  # type: ignore

    async def __aexit__(
        self,
        exception_type: Type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None
    ) -> None:
        if self.websocket_session and not self.websocket_session.closed:
            await self.websocket_session.close()

    async def get_stream(
        self,
        query: str | Sequence[str],
        url: str,
        headers: dict[str, str] | None = None
    ) -> AsyncIterator[str]:
        assert self.websocket_session is not None, \
            'API has to be used as a context manager only'

        if isinstance(query, str):
            query = [query]

        async with self.websocket_session.ws_connect(
            url, headers=headers, ssl_context=self.ssl_context
        ) as websocket:
            for message in query:
                await websocket.send_str(message)
            self.logger.debug('Message sent to websocket: %s', query)

            async for response in websocket:  # type: WSMessage
                if response.type == WSMsgType.TEXT:
                    self.logger.debug('Got message: %s', response.data)

                    yield response.data



from __future__ import annotations

import json

from base64 import b64encode
from collections.abc import Iterable, Sequence
from datetime import datetime, date, time
from io import BytesIO
from typing import Any
from urllib.parse import quote

from lxml.html import parse

from tradernet.core import TraderNetCore
from tradernet.string_utils import extract_zip


class TraderNetAPI(TraderNetCore):
    """
    Client methods to interact TraderNet API.
    """
    def new_user(
        self,
        login: str,
        reception: str | int,
        phone: str,
        lastname: str,
        firstname: str,
        password: str | None = None,
        utm_campaign: str | None = None,
        tariff: int | None = None
    ) -> dict[str, str | int]:
        """
        Creating a new user.

        Parameters
        ----------
        login : str
            A login.
            A password.
        reception : str | int
            A reception number.
        phone : str | None
            User's phone no.
        lastname : str | None
            User's last name.
        firstname : str | None
            User's first name.
        password : str | None
            User's password. If None, it will be generated automatically.
        utm_campaign : str | None
            Referral link. This field is used if a new user is created after
            receiving a referral link.
        tariff : int | None
            Selected rate ID. Optional parameter. During the registration, you
            may immediately assign the desired rate ID.

        Returns
        -------
        dict[str, str | int]
            A dictionary with the following keys: 'clientId', 'userId'.

        Notes
        -----
        https://tradernet.ru/tradernet-api/primary-registration
        """
        cmd = 'registerNewUser'
        params = {
            'login': login,
            'pwd': password,
            'reception': str(reception),
            'phone': phone,
            'lastname': lastname,
            'firstname': firstname,
            'tariff_id': tariff,
            'utm_campaign': utm_campaign
        }
        return self.plain_request(cmd, params)

    def change_phone(self, phone: str) -> dict[str, Any]:
        """
        Changing a phone number.

        Parameters
        ----------
        phone : str
            A phone number.

        Returns
        -------
        dict[str, Any]
            A dictionary with the following keys: 'phoneId'.

        Notes
        -----
        https://tradernet.ru/tradernet-api/check-phone
        """
        cmd = 'checkPhone'
        params = {'phone': phone}
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd, params)

    def check_sms(self, phone_id: int, code: str) -> dict[str, Any]:
        """
        Method for checking the validity of the code sent in SMS using the
        method checkPhone.

        Parameters
        ----------
        phone_id : int
            A phone ID.
        code : str
            A SMS code.

        Returns
        -------
        dict[str, Any]
            A dictionary with the following keys: 'success'.

        Notes
        -----
        https://tradernet.ru/tradernet-api/check-phone
        """
        cmd = 'checkSms'
        params = {'phoneId': phone_id, 'code': code}
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd, params)

    def get_tariffs_list(self) -> dict[str, Any]:
        """
        Get a list of available tariffs.

        Returns
        -------
        dict[str, Any]
            Tariffs list.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-list-tariff
        """
        cmd = 'GetListTariffs'
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd)

    def select_tariff(self, tariff_id: int) -> dict[str, Any]:
        """
        Selecting a tariff.

        Parameters
        ----------
        tariff_id : int
            A tariff ID.

        Returns
        -------
        dict[str, Any]
            A dictionary with the following keys: 'added'.

        Notes
        -----
        https://tradernet.ru/tradernet-api/select-tariff
        """
        cmd = 'selectTariff'
        params = {'tariff_id': tariff_id}
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd, params)

    def get_agreement(self) -> bytes:
        """
        Receiving application for joining in PDF format.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-agreement-pdf
        """
        cmd = 'getAgreementPdf'
        if not self._session_id:
            self.get_authorized()

        message: dict[str, Any] = {'cmd': cmd, 'SID': self._session_id}
        url = f'{self.url}/api'
        query = {'q': json.dumps(message)}

        response = self.request('get', url, params=query)
        return response.content

    def send_agreement(
        self,
        agreement: bytes,
        check: int | None
    ) -> dict[str, Any]:
        """
         Uploading a signed application for joining in PDF format.

        Parameters
        ----------
        agreement : bytes
            An application for joining in PDF format.
        check : int | None
            1 - Mandatory check. More details can be found on the page:
            https://tradernet.ru/tradernet-api/special-files-list

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-agreement-pdf
        """
        cmd = 'uploadFiles'
        agreement = b64encode(agreement)

        params = {
            'name': 'anketa_files_signed_agreement',
            'extension': 'pdf',
            'img': agreement,
            'check': check
        }

        return self.authorized_request(cmd, params, version=None)

    def check_missing_fields(self, step: int, office: str) -> dict[str, Any]:
        """
        Checking missing (blank) fields. If any fields are missing, they will
        be specified in the `not_completed` parameter, along with a
        description.

        Parameters
        ----------
        step : int
            A step number.
        office : str
            An office name.

        Notes
        -----
        https://tradernet.ru/tradernet-api/check-step
        """
        cmd = 'checkStep'
        params = {'step': step, 'office': office}
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd, params)

    def get_profile_fields(self, reception: int) -> dict[str, Any]:
        """
        Obtaining profile fields for different offices.

        Parameters
        ----------
        reception : int
            A reception number.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-anketa-fields
        """
        cmd = 'getAnketaFields'
        params = {'anketa_for_reception': reception}
        return self.plain_request(cmd, params)

    def user_info(self) -> dict[str, Any]:
        """
        Obtaining user information.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-user-info
        """
        cmd = 'GetAllUserTexInfo'
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd)

    def sign_application(self) -> dict[str, Any]:
        """
        Signing the application by SMS sent to the phone specified by the new
        user.

        Notes
        -----
        https://tradernet.ru/tradernet-api/sign-anketa-electronically
        """
        cmd = 'signAnketaElectronically'
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd)

    def verify_application_signature(
        self,
        phone_id: str,
        code: str
    ) -> dict[str, Any]:
        """
        Confirming the signature via SMS, received to the client's phone number
        using the method `sign_application`.

        Parameters
        ----------
        phone_id : str
            A phone ID.
        code : str
            A SMS code.

        Notes
        -----
        https://tradernet.ru/tradernet-api/check-anketa-electronically-sign-sms-code
        """
        cmd = 'checkAnketaElectronicallySignSmsCode'
        params = {'phoneId': phone_id, 'code': code}
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd, params)

    def get_user_data(self) -> dict[str, Any]:
        """
        Getting initial user data from the server - orders, portfolio, markets,
        open sessions, etc.

        Notes
        -----
        https://tradernet.ru/tradernet-api/auth-get-opq
        """
        cmd = 'getOPQ'
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd)

    def get_market_status(
        self,
        market: str = '*',
        mode: str | None = None
    ) -> dict[str, Any]:
        """
        Obtaining information about market statuses and operation.

        Parameters
        ----------
        market : str
            A market code (briefName).
        mode : str | None
            Request mode: demo. If the parameter is not specified, the market
            statuses for real users will be displayed.

        Notes
        -----
        https://tradernet.ru/tradernet-api/market-status
        """
        cmd = 'getMarketStatus'
        params = {'market': market}
        if mode:
            params['mode'] = mode
        return self.plain_request(cmd)

    def security_info(self, symbol: str, sup: bool = True) -> dict[str, Any]:
        """
        Getting info on a specific symbol.

        Parameters
        ----------
        symbol : str
            A TraderNet symbol.
        sup : bool
            IMS and trading system format.

        Returns
        -------
        result : dict
            A dictionary of symbol info.

        Notes
        -----
        https://tradernet.ru/tradernet-api/quotes-get-info
        """
        cmd = 'getSecurityInfo'
        params = {'ticker': symbol, 'sup': sup}
        return self.authorized_request(cmd, params=params, version=1)

    def get_options(
        self,
        underlying: str,
        exchange: str
    ) -> list[dict[str, str]]:
        """
        Downloading a list of active options by the underlying asset and
        exchange.

        Parameters
        ----------
        underlying : str
            The underlying symbol.
        exchange : str
            A venue options traded.

        Returns
        -------
        list[dict[str, str]]
            List of very basic properties of options.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-options-by-mkt
        """
        cmd = 'getOptionsByMktNameAndBaseAsset'
        params = {'base_contract_code': underlying, 'ltr': exchange}
        return self.plain_request(cmd, params)

    def get_most_traded(
        self,
        instrument_type: str = 'stocks',
        exchange: str = 'usa',
        gainers: bool = True,
        limit: int = 10
    ) -> dict[str, Any]:
        """
        Getting a list of the most traded securities or a list of the fastest
        growing stocks (for a year).

        Parameters
        ----------
        instrument_type : str
            Instrument type.
        exchange : str
            Stock exchanges. Possible values: 'usa', 'europe', 'ukraine',
            'currencies'.
        gainers : bool
            True: top fastest-growing, False: top by trading volume.
        limit : int
            Number of instruments displayed.

        Notes
        -----
        https://tradernet.ru/tradernet-api/quotes-get-top-securities
        """
        cmd = 'getTopSecurities'
        params = {
            'type': instrument_type,
            'exchange': exchange,
            'gainers': int(gainers),
            'limit': limit
        }
        return self.plain_request(cmd, params)

    def export_securities(
        self,
        symbols: str | Sequence[str],
        fields: list[str] | None = None
    ) -> list[dict[str, Any]]:
        """
        Exporting securities data from TraderNet.

        Parameters
        ----------
        symbols : str | Sequence[str]
            A symbol or a list of symbols.
        fields : list[str] | None, optional
            Limiting fields, by default None which means all fields.

        Returns
        -------
        list[dict[str, Any]]
            A list of dictionaries with security data.

        Notes
        -----
        https://tradernet.ru/tradernet-api/quotes-get
        """
        if isinstance(symbols, str):
            symbols = [symbols]

        params = {'tickers': ' '.join(symbols)}
        if fields:
            params['params'] = ' '.join(fields)

        url = f'{self.url}/securities/export'
        response = self.request(
            'get', url, headers=self.HEADERS, params=params
        )
        return response.json()

    def get_candles(
        self,
        symbol: str,
        start: datetime = datetime(2010, 1, 1),
        end: datetime = datetime.utcnow(),
        timeframe: int = 86400
    ) -> dict[str, Any]:
        """
        Getting historical data of a symbol.

        Parameters
        ----------
        symbol : str
            A symbol name on TraderNet.
        start : datetime
            The first date of the period market data to be acquired within.
        end : datetime
            The last date of the period.
        timeframe : int
            Timeframe of candles in seconds. Default is 86400 corresponding to
            day candles. -1 value of the parameter indicating that traders are
            required.

        Returns
        -------
        result : dict
            A dictionary of historical information of the symbol.

        Notes
        -----
        https://tradernet.ru/tradernet-api/quotes-get-hloc
        https://tradernet.ru/tradernet-api/get-trades
        """
        cmd = 'getHloc'
        params = {
            'id': symbol,
            'count': -1,
            'timeframe': int(timeframe / 60),
            'date_from': start.strftime('%d.%m.%Y %H:%M'),
            'date_to': end.strftime('%d.%m.%Y %H:%M'),
            'intervalMode': 'ClosedRay'
        }
        return self.authorized_request(cmd, params=params)

    def get_trades_history(
        self,
        start: date = date(1970, 1, 1),
        end: date = date.today(),
        trade_id: int | None = None,
        limit: int | None = None,
        symbol: str | None = None,
        currency: str | None = None,
        reception: int | str | None = None
    ) -> dict[str, Any]:
        """
        Getting a list of trades.

        Parameters
        ----------
        start : date
            Period start date.
        end : date
            Period end date.
        trade_id : int | None
            From which Trade ID to start retrieving report data.
        limit : int | None
            Number of trades. If 0 or no parameter is specified - then all
            trades.
        symbol : str | None
            A symbol.
        currency : str | None
            Base currency or quote currency.
        reception : int | str | None
            Office ID.

        Returns
        -------
        result : dict
            A dictionary of trades.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-trades-history
        """
        cmd = 'getTradesHistory'
        params = {
            'beginDate': str(start),
            'endDate': str(end),
            'tradeId': trade_id,
            'max': limit,
            'nt_ticker': symbol,
            'curr': currency,
            'reception': reception
        }
        if not self._session_id:
            self.get_authorized()
        return self.plain_request(cmd, params)

    def find_symbol(
        self,
        symbol: str,
        exchange: str | None = None
    ) -> dict[str, Any]:
        """
        Stock symbols search.

        Parameters
        ----------
        symbol : str
            A symbol name.
        exchange : str, optional
            Refbook name.

        Returns
        -------
        result : dict
            A dictionary of symbols, max 30.

        Notes
        -----
        https://tradernet.ru/tradernet-api/quotes-finder
        """
        cmd = 'tickerFinder'
        params = {'text': f'{symbol}@{exchange}' if exchange else symbol}
        return self.plain_request(cmd, params)

    def get_news(
        self,
        query: str,
        symbol: str | None = None,
        story_id: str | None = None,
        limit: int = 30
    ) -> dict[str, Any]:
        """
        Getting news on securities.

        Parameters
        ----------
        query : str
            Can be ticker or any word.
        symbol : str | None
            If parameter symbol is set, `query` will be ignored and newsfeed
            will consist only of stories solely based on mentioned symbol.
        story_id : str | None
            If parameter story_id is set, `query` and `symbol` parameters will
            be ignored and news feed will consist only of the story with
            required storyId.
        limit : int
            Max number of news, 30 by default.

        Notes
        -----
        https://tradernet.ru/tradernet-api/quotes-get-news
        """
        cmd = 'getNews'
        params = {
            'searchFor': query,
            'ticker': symbol,
            'storyId': story_id,
            'limit': limit
        }
        if not self._session_id:
            self.get_authorized()
        return self.plain_request(cmd, params)

    def get_all(
        self,
        filters: dict[str, Any] | None = None,
        show_expired: bool = False
    ) -> list[dict[str, Any]]:
        """
        Getting information on securities.

        Parameters
        ----------
        filters : dict, optional
            Field names and their values.
        show_expired : bool, optional
            Getting expired symbols or not. None value sets the field
            'maturity_d' to null.

        Returns
        -------
        result : dict
            A dictionary of symbols.

        Notes
        -----
        https://tradernet.ru/tradernet-api/securities

        Examples
        --------
        >>> self.get_all(filters={'mkt_short_code': 'FIX',
        >>>                       'instr_type_c': '1'})
        dict(...)

        This command is to obtain all stocks of the FIX (American composite)
        venue.

        >>> self.get_all(filters={'mkt_short_code': 'ICE',
        >>>                       'instr_type_c': '4'})
        dict(...)

        And this one obtains all active options from the ICE exchange.
        """
        # TODO: implement wildcards
        if filters:
            filter_list: list[dict[str, str | int]] = [
                {
                    'field': field,
                    'operator': 'eq',
                    'value': quote(value) if isinstance(value, str) else value
                }
                for field, value in filters.items()
            ]
        else:
            filter_list = []

        if not show_expired and (not filters or 'istrade' not in filters):
            filter_list.append(
                {'field': 'istrade', 'operator': 'eq', 'value': 1}
            )

        symbols: list[dict[str, Any]] = []
        skip = 0
        while True:
            message = {
                'skip': skip,
                'take': self.CHUNK_SIZE,
                'filter': {'filters': filter_list}
            }
            self.logger.debug('Message string: %s', message)
            query = self.http_build_query(message)
            response = self.request(
                'get',
                f'{self.url}/securities/ajax-get-all-securities',
                params=query
            )

            symbols_chunk = json.loads(response.text)['securities']
            symbols += symbols_chunk

            if len(symbols_chunk) < self.CHUNK_SIZE:
                self.logger.debug('The final chunk of symbols is received')
                return symbols

            skip += self.CHUNK_SIZE

    def account_summary(self) -> dict[str, Any]:
        """
        Getting summary of own account.

        Returns
        -------
        result : dict
            A dictionary of all positions, active orders, etc.

        Notes
        -----
        https://tradernet.ru/tradernet-api/portfolio-get-changes
        """
        cmd = 'getPositionJson'
        return self.authorized_request(cmd)

    def get_price_alerts(self, symbol: str | None = None) -> dict[str, Any]:
        """
        Getting a list of price alerts.

        Parameters
        ----------
        symbol : str | None, optional
            Symbol to get alerts for.

        Returns
        -------
        result : dict
            A dictionary of alerts.

        Notes
        -----
        https://tradernet.ru/tradernet-api/alerts-get-list
        """
        cmd = 'getAlertsList'
        params: dict[str, str] = {}
        if symbol:
            params['ticker'] = symbol

        if not self._session_id:
            self.get_authorized()
        return self.plain_request(cmd, params)

    def add_price_alert(
        self,
        symbol: str,
        price: int | float | str | Iterable[int | float | str],
        trigger_type: str = 'crossing',
        quote_type: str = 'ltp',
        send_to: str = 'email',
        frequency: int = 0,
        expire: int = 0
    ) -> dict[str, Any]:
        """
        Adding a price alert.

        Parameters
        ----------
        symbol : str
            Symbol to add alert for.
        price : float | list[float]
            Price of the alert activation.
        trigger_type : str, optional
            Trigger method.
        quote_type : str, optional
            Type of the price underlying the alert calculation. Possible
            values: 'ltp', 'bap', 'bbp', 'op', 'pp'.
        send_to : str, optional
            Type of notification. Possible values: 'email', 'sms', 'push',
            'all', by default 'email'.
        frequency : int, optional
            Frequency.
        expire : int, optional
            Alert period.

        Returns
        -------
        result : dict
            Addition result.

        Notes
        -----
        https://tradernet.ru/tradernet-api/alerts-add
        """
        cmd = 'addPriceAlert'
        if not isinstance(price, Iterable):
            price = [str(price)]
        else:
            price = [*map(str, price)]
        params = {
            'ticker': symbol,
            'price': price,
            'trigger_type': trigger_type,
            'quote_type': quote_type,
            'notification_type': send_to,
            'alert_period': frequency,
            'expire': expire
        }
        if not self._session_id:
            self.get_authorized()
        return self.plain_request(cmd, params)

    def delete_price_alert(self, alert_id: int) -> dict[str, Any]:
        """
        Deleting a price alert.

        Parameters
        ----------
        alert_id : int
            Alert ID.

        Returns
        -------
        result : dict
            Deletion result.

        Notes
        -----
        https://tradernet.ru/tradernet-api/alerts-delete
        """
        cmd = 'addPriceAlert'
        params = {'id': alert_id, 'del': True}
        if not self._session_id:
            self.get_authorized()
        return self.plain_request(cmd, params)

    def get_requests_history(
        self,
        doc_id: int | None = None,
        exec_id: int | None = None,
        start: date = datetime(1970, 1, 1),
        end: date = datetime.utcnow(),
        limit: int | None = None,
        offset: int | None = None,
        status: int | None = None
    ) -> dict[str, Any]:
        """
        Receiving clients' requests history.

        Parameters
        ----------
        doc_id : int | None, optional
            Request type ID.
        exec_id : int | None, optional
            Order ID.
        start : date, optional
            Period start date.
        end : date, optional
            Period end date.
        limit : int | None, optional
            Number of orders displayed in the list.
        offset : int | None, optional
            Step of the list of displayed requests.
        status : int | None, optional
            Requests statuses: 0 - draft request; 1 - in process of execution;
            2 - request is rejected; 3 - request is executed.

        Returns
        -------
        result : dict
            Clients' requests for the specified period.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-client-cps-history
        """
        cmd = 'getClientCpsHistory'
        params: dict[str, str | int] = {
            'date_from': start.strftime('%Y-%m-%dT%H:%M:%S'),
            'date_to': end.strftime('%Y-%m-%dT%H:%M:%S')
        }
        if doc_id:
            params['cpsDocId'] = doc_id
        if exec_id:
            params['id'] = exec_id
        if limit:
            params['limit'] = limit
        if offset:
            params['offset'] = offset
        if status:
            params['cps_status'] = status
        return self.authorized_request(cmd, params)

    def get_order_files(
        self,
        order_id: int | None,
        internal_id: int | None
    ) -> dict[str, Any]:
        """
        Receiving order files.

        Parameters
        ----------
        order_id : int | None, optional
            Order ID. May be not used if the draft order ID is known
            (internal_id).
        internal_id : int | None, optional
            Draft order number. Used when known, or if the order has the draft
            status and has not yet been assigned the main ID.

        Returns
        -------
        result : dict
            Order files.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-cps-files
        """
        cmd = 'getCpsFiles'
        params: dict[str, int] = {}
        if internal_id:
            params['internal_id'] = internal_id
        elif order_id:
            params['id'] = order_id
        else:
            raise ValueError(
                'Either order_id or internal_id must be specified'
            )
        return self.authorized_request(cmd, params)

    def get_broker_report(
        self,
        start: date = date(1970, 1, 1),
        end: date = date.today(),
        period: time = time(23, 59, 59),
        data_block_type: str | None = 'account_at_end'
    ) -> dict[str, Any]:
        """
        Getting the broker's report using software methods.

        Parameters
        ----------
        start : date, optional
            Period start date.
        end : date, optional
            Period end date.
        period : time, optional
            Time cut maybe 23:59:59 or 08:40:00.
        data_block_type : str | None, optional
            Data block from the report.

        Returns
        -------
        dict[str, Any]
            Broker's report.
        """
        cmd = 'getBrokerReport'
        params: dict[str, str | None] = {
            'date_start': str(start),
            'date_end': str(end),
            'time_period': period.strftime('%H:%M:%S'),
            'format': 'json',
            'type': data_block_type
        }
        return self.authorized_request(cmd, params)

    def get_depository_report(
        self,
        start: date = datetime(1970, 1, 1),
        end: date = date.today(),
        period: time = time(23, 59, 59),
        file_format: str = 'pdf',
        data_block_type: str | None = 'account_at_end',
        encoded_result: bool = True
    ) -> dict[str, Any]:
        """
        Getting the depository's report using software methods.

        Parameters
        ----------
        start : date, optional
            Period start date.
        end : date, optional
            Period end date.
        period : time, optional
            Time cut maybe 23:59:59 or 08:40:00.
        file_format : str, optional
            Report in the selected format. Acceptable formats: 'json', 'html',
            'xml', 'xls', 'pdf'.
        data_block_type : str | None, optional
            Data block from the report.
        encoded_result : bool, optional
            Retrieves the report as a JSON array containing the encoded file in
            section file base64.

        Returns
        -------
        dict[str, Any]
            Depository's report.
        """
        cmd = 'getDepositoryReport'
        params: dict[str, int | str | None] = {
            'date_start': str(start),
            'date_end': str(end),
            'time_period': period.strftime('%H:%M:%S'),
            'format': file_format,
            'type': data_block_type,
            'encoded_result': int(encoded_result)
        }
        return self.authorized_request(cmd, params)

    def symbol(self, symbol: str, lang: str = 'en') -> dict[str, Any]:
        """
        A method for obtaining information on a given security.

        Parameters
        ----------
        symbol : str
            A TraderNet symbol.
        lang : str
            Language, two letters.

        Returns
        -------
        result : dict
            A dictionary of symbol info.

        Notes
        -----
        https://tradernet.ru/tradernet-api/shop-get-stock-data
        """
        cmd = 'getStockData'
        params = {'ticker': symbol, 'lang': lang}
        return self.authorized_request(cmd, params=params, version=1)

    def symbols(self, exchange: str | None = None) -> dict[str, Any]:
        """
        Receiving completed lists of securities.

        Parameters
        ----------
        exchange : str, optional
            Optional parameter that allows to get data from NYSE and NASDAQ or
            Moscow Exchange. May accept the value usa, russia.

        Returns
        -------
        result : dict
            A dictionary of exchanges and symbols.

        Notes
        -----
        https://tradernet.ru/tradernet-api/get-ready-list
        """
        cmd = 'getReadyList'
        if exchange:
            params: dict[str, str] | None = {'mkt': exchange}
        else:
            params = None
        return self.authorized_request(cmd, params=params, version=1)

    def refbooks(
        self,
        reference_date: date | None = date.today()
    ) -> list[str]:
        """
        Getting the list of available reference books.

        Parameters
        ----------
        reference_date : date
            Reference date of the book.

        Returns
        -------
        result : list
            The list of reference books for a specific date found.
        """
        url = f'{self.url}/refbooks/{reference_date}'
        page = self.request('get', url).content
        doc = parse(BytesIO(page)).getroot()
        result = [div.text_content().rsplit('.', 2)[0]
                  for div in doc.cssselect('a')]
        result.remove('')
        return result

    def get_refbook(
        self,
        name: str | None = 'all',
        reference_date: date | None = date.today()
    ) -> list[dict[str, Any]]:
        """
        Downloading and processing a particular reference book.

        Parameters
        ----------
        name : str
            The name of the book.
        reference_date : date
            Reference date of the book.

        Returns
        -------
        result : list
            The list of all instruments with their properties found in the
            reference book.
        """
        if name == 'all':
            self.logger.warning('Downloading all symbols may take a while!')
        archive = f'{name}.json.zip'
        uri = f'{self.url}/refbooks/{reference_date}/{archive}'
        # Download and unzip
        archive_json = extract_zip(self.request('get', url=uri).content)
        self.logger.debug(
            'Files in the archive: %s', ', '.join(archive_json.keys())
        )

        assert len(archive_json) == 1, 'More than one file in archive!'
        return json.loads(archive_json[f'{name}.json'])

    def corporate_actions(
        self,
        reception: int = 35
    ) -> list[dict[str, Any]]:
        """
        Getting planned corporate actions for a certain office.

        Parameters
        ----------
        reception : 35
            Office number.

        Returns
        -------
        result : list
            Expected corporate actions.
        """
        cmd = 'getPlannedCorpActions'
        params = {'reception': reception}
        return self.plain_request(cmd, params)



from __future__ import annotations

from collections.abc import Callable
from datetime import datetime
from typing import Any, ClassVar

from tradernet.core import TraderNetCore


class Trading(TraderNetCore):
    DURATION: ClassVar[dict[str, int]] = {
        'day': 1,  # The order will be valid until the end of the trading day.
        'ext': 2,  # Extended day order.
        'gtc': 3   # A.k.a. "Good Till Cancelled"
    }

    def open_security_session(
        self,
        *args: bool,
        **kwargs: bool
    ) -> Callable[[str], dict[str, Any]]:
        """
        Opening a new security session.

        Returns
        -------
        result : Callable[[str], dict]
            A function that accepts a security code and returns an answer
            containing login information.

        Notes
        -----
        Use it as follows:
        >>> trade = Trading.from_config('tradernet.ini')
        >>> opening = trade.open_security_session(False, True)  # Send to app
        >>> opening(123456)  # Push you received
        """
        self.send_security_sms(*args, **kwargs)
        return lambda token: self.open_with_sms(str(token))

    def buy(
        self,
        symbol: str,
        quantity: int = 1,
        price: float | None = None,
        duration: str = 'day',
        use_margin: bool = True
    ) -> dict[str, Any]:
        cmd = 'putTradeOrder'
        params = {
            'instr_name': symbol,
            'action_id': 2 if use_margin else 1,
            'order_type_id': 2 if price else 1,
            'qty': quantity,
            'limit_price': price,
            'expiration_id': self.DURATION[duration.lower()]
        }
        return self.authorized_request(cmd, params, version=2)

    def sell(
        self,
        symbol: str,
        quantity: int = 1,
        price: float | None = None,
        duration: str = 'day',
        use_margin: bool = True
    ) -> dict[str, Any]:
        cmd = 'putTradeOrder'
        params = {
            'instr_name': symbol,
            'action_id': 4 if use_margin else 3,
            'order_type_id': 2 if price else 1,
            'qty': quantity,
            'limit_price': price,
            'expiration_id': self.DURATION[duration.lower()]
        }
        return self.authorized_request(cmd, params, version=2)

    def stop(self, symbol: str, price: float) -> dict[str, Any]:
        cmd = 'putStopLoss'
        params = {
            'instr_name': symbol,
            'stop_loss': price
        }
        return self.authorized_request(cmd, params, version=2)

    def trailing_stop(self, symbol: str, percent: int = 1) -> dict[str, Any]:
        cmd = 'putStopLoss'
        params = {
            'instr_name': symbol,
            'stop_loss_percent': percent,
            'stoploss_trailing_percent': percent
        }
        return self.authorized_request(cmd, params, version=2)

    def cancel(self, order_id: int) -> dict[str, Any]:
        cmd = 'delTradeOrder'
        params = {'order_id': order_id}
        return self.authorized_request(cmd, params, version=2)

    def get_placed(self, active: bool = True) -> dict[str, Any]:
        """
        Getting a list of orders in the current period.

        Parameters
        ----------
        active : bool, optional
            Show only active orders.

        Returns
        -------
        result : dict
            A dictionary of orders.

        Notes
        -----
        https://tradernet.ru/tradernet-api/orders-get-current-history
        """
        cmd = 'getNotifyOrderJson'
        params = {'active_only': int(active)}
        return self.authorized_request(cmd, params)

    def get_historical(
        self,
        start: datetime = datetime(1970, 1, 1),
        end: datetime = datetime.utcnow()
    ) -> dict[str, Any]:
        """
        Getting a list of orders in the period.

        Parameters
        ----------
        start : datetime, optional
            Period start date.
        end : datetime, optional
            Period end date.

        Returns
        -------
        result : dict
            A dictionary of orders.

        Notes
        -----
        https://tradernet.ru/tradernet-api/orders-get-history
        """
        cmd = 'getOrdersHistory'
        params = {
            'from': start.strftime('%Y-%m-%dT%H:%M:%S'),
            'till': end.strftime('%Y-%m-%dT%H:%M:%S')
        }
        if not self._session_id:
            self.get_authorized()
        return self.plain_request(cmd, params)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from datetime import date, datetime
from decimal import Decimal
from functools import total_ordering

from tradernet.symbols.option_properties import OptionProperties


@total_ordering
class BaseOptionSymbol(metaclass=ABCMeta):
    """
    Parsing class for option names.
    """
    __slots__ = ('symbol', '__properties')

    def __init__(self, symbol: str) -> None:
        self.symbol = symbol
        self.__properties = self.decode_notation(self.symbol)

    def __repr__(self) -> str:
        if self.right == -1:
            right = 'Put'
        else:
            right = 'Call'
        underlying = self.underlying
        return f'{underlying} @ {self.strike} {right} {self.maturity_date}'

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, BaseOptionSymbol), \
            'Trying to compare different types'

        return (
            self.underlying, self.maturity_date, self.strike, self.right
        ) == (
            other.underlying, other.maturity_date, other.strike, other.right
        )

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __lt__(self, other: object) -> bool:
        assert isinstance(other, BaseOptionSymbol), \
            'Trying to compare different types'

        return (
            self.underlying, self.maturity_date, self.strike, self.right
        ) < (
            other.underlying, other.maturity_date, other.strike, other.right
        )

    @property
    def ticker(self) -> str:
        return self.__properties.ticker

    @property
    def location(self) -> str | None:
        return self.__properties.location

    @property
    def right(self) -> int:
        return self.__properties.right

    @property
    def strike(self) -> Decimal:
        return self.__properties.strike

    @property
    def maturity_date(self) -> date:
        return self.__properties.maturity_date

    @property
    def symbolic_expiration(self) -> str:
        return self.__properties.symbolic_expiration

    @property
    def underlying(self) -> str:
        """
        Actually, this is only a guess of the underlying symbol.

        Returns
        -------
        str
            Underlying symbol.
        """
        if not self.location:
            return self.ticker
        return f'{self.ticker}.{self.location}'

    @property
    def symbolic_right(self) -> str:
        return 'C' if self.right == 1 else 'P'

    @staticmethod
    def numeric_right(is_call: bool) -> int:
        assert isinstance(is_call, bool), 'Wrong type of argument'
        return (-1)**(not is_call)

    def osi(self) -> str:
        """
        Converting option name to OSI format.

        Returns
        -------
        str
            Resulting name.
        """
        expiration = self.maturity_date.strftime('%y%m%d')

        strike = str(self.strike).split('.')
        dollar = strike[0].zfill(5)
        if len(strike) == 2:
            cent = strike[1]
        else:
            cent = '0'
        cent = cent.zfill(3)

        return f'{self.ticker}{expiration}{self.symbolic_right}{dollar}{cent}'

    @staticmethod
    @abstractmethod
    def decode_date(symbolic_date: str) -> date:
        """
        Decoding date from string.

        Parameters
        ----------
        symbolic_date : str
            String with date.

        Returns
        -------
        date
            Decoded date.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def encode_date(conventional_date: str | date | datetime) -> str:
        """
        Encoding date to a concrete notation.

        Parameters
        ----------
        conventional_date : str | date | datetime | Date
            A date.

        Returns
        -------
        str
            Converted date.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def decode_notation(symbol: str) -> OptionProperties:
        """
        Decoding a particular option notation.
        """
        raise NotImplementedError



from __future__ import annotations

from collections.abc import Callable, MutableMapping
from hmac import new as hmac_new
from json import dumps as json_dumps
from io import BytesIO
from typing import Any, Union
from zipfile import ZipFile

from mypy_extensions import trait


JSON_VALUE_TYPE = Union[str, int, float, bool, None]


def extract_zip(content: bytes) -> dict[str, str]:
    """
    Extracting a zipped file into a dictionary having the structure
    "name->content".
    Parameters
    ----------
    content : bytes
        Byte-representation of an archive.
    Returns
    -------
    dict[str, str]
        Resulting dictionary.
    """
    with ZipFile(BytesIO(content)) as zip_file:
        return {name: zip_file.read(name).decode()
                for name in zip_file.namelist()}


@trait
class StringUtils:
    @staticmethod
    def simple_list(raw_list: list[Any]) -> str:
        """
        ['a', 'b'] => "['a', 'b']"
        Parameters
        ----------
        raw_list : list
            Any list.
        Returns
        -------
        result : str
            A string looking like a normal Python list with square brackets.
        """
        string_values = map(lambda x: f"'{x}'", raw_list)
        string_list = ', '.join(string_values)
        return f'[{string_list}]'

    @staticmethod
    def flatten_list(
        raw_list: list[Any],
        parent_key: str = ''
    ) -> dict[str, Any]:
        """
        ['a', ['b', 'c']] => {'[0]': 'a', '[1][0]': 'b', '[1][1]': 'c'}
        """
        items: list[tuple[str, Any]] = []
        for key, value in enumerate(raw_list):
            new_key = f'{parent_key}[{key}]' if parent_key else f'[{key}]'
            if isinstance(value, MutableMapping):
                items.extend(StringUtils.flatten_dict(value, new_key).items())
            elif isinstance(value, list):
                items.extend(StringUtils.flatten_list(value, new_key).items())
            else:
                items.append((new_key, value))
        return dict(items)

    @staticmethod
    def flatten_dict(
        dictionary: MutableMapping[str, Any],
        parent_key: str = ''
    ) -> dict[str, Any]:
        """
        {'a': 1, 'b': {'c': 2}} => {'[a]': 1, '[b][c]': 2}
        """
        items: list[tuple[str, Any]] = []
        for key, value in dictionary.items():
            new_key = f'{parent_key}[{key}]' if parent_key else f'[{key}]'
            if isinstance(value, MutableMapping):
                items.extend(StringUtils.flatten_dict(value, new_key).items())
            elif isinstance(value, list):
                items.extend(StringUtils.flatten_list(value, new_key).items())
            else:
                items.append((new_key, value))
        return dict(items)

    @staticmethod
    def str_from_list(items: list[Any]) -> str:
        strings: list[str] = []
        for value in items:
            if isinstance(value, MutableMapping):
                value = StringUtils.str_from_dict(value)
            elif isinstance(value, list):
                value = StringUtils.str_from_list(value)
            else:
                value = str(value)
            strings.append(value)
        return StringUtils.simple_list(strings)

    @staticmethod
    def str_from_dict(dictionary: MutableMapping[str, Any]) -> str:
        """
        {'key1': value1, 'key2': value2} => 'key1=value1&key2=value2'
        """
        strings: list[str] = []
        for key, value in dictionary.items():
            if isinstance(value, MutableMapping):
                value = StringUtils.str_from_dict(value)
            elif isinstance(value, list):
                value = StringUtils.simple_list(value)
            else:
                value = str(value)
            strings.append(f'{key}={value}')
        return '&'.join(sorted(strings))

    @staticmethod
    def stringify(items: list[Any]) -> str:
        return json_dumps(items, separators=(',', ':'))

    @staticmethod
    def sign(
        key: str,
        message: str = '',
        algorithm_name: str = 'sha256'
    ) -> str:
        """
        Signing a message with a key.

        Parameters
        ----------
        key : str
            A private key.
        message : str, optional
            A message to be sign, by default ''.
        algorithm_name : str, optional
            The name of an algorithm for signing, by default sha256.

        Returns
        -------
        str
            Signed message.
        """
        return hmac_new(
            key.encode(), message.encode(), digestmod=algorithm_name
        ).hexdigest()

    @staticmethod
    def apply_to_json(
        function: Callable[[str | int], JSON_VALUE_TYPE],
        source: MutableMapping[str, Any] | list[Any]
    ) -> dict[str, Any] | list[Any]:
        if isinstance(source, MutableMapping):
            return StringUtils.apply_to_dict(function, source)
        if isinstance(source, list):
            return StringUtils.apply_to_list(function, source)
        raise TypeError('Source must be a dict or a list.')

    @staticmethod
    def apply_to_list(
        function: Callable[[str | int], JSON_VALUE_TYPE],
        items: list[Any]
    ) -> list[Any]:
        new_items: list[Any] = []
        for value in items:
            if isinstance(value, MutableMapping):
                value = StringUtils.apply_to_dict(function, value)
            elif isinstance(value, list):
                value = StringUtils.apply_to_list(function, value)
            else:
                value = function(value)
            new_items.append(value)
        return new_items

    @staticmethod
    def apply_to_dict(
        function: Callable[[str | int], JSON_VALUE_TYPE],
        items: MutableMapping[str, Any]
    ) -> dict[str, Any]:
        new_items: dict[str, Any] = {}
        for key, value in items.items():
            if isinstance(value, MutableMapping):
                value = StringUtils.apply_to_dict(function, value)
            elif isinstance(value, list):
                value = StringUtils.apply_to_list(function, value)
            else:
                value = function(value)
            new_items[key] = value
        return new_items



from __future__ import annotations

from configparser import ConfigParser
from copy import deepcopy
from datetime import datetime
from json import dumps as json_dumps
from logging import getLogger
from typing import Any, ClassVar, Type, TypeVar

from tradernet.netutils import NetUtils


Self = TypeVar('Self', bound='TraderNetCore')


class TraderNetCore(NetUtils):
    """
    Core tools to interact TraderNet API.

    Parameters
    ----------
    public : str, optional
        A TraderNet public key.
    private: str, optional
        A TraderNet private key.
    login : str, optional
        A TraderNet login.
    password : str, optional
        A password for the login.

    Attributes
    ----------
    logger : Logger
        Handling errors and warnings.
    """
    DOMAIN: ClassVar[str] = 'tradernet.com'  # TraderNet server
    SESSION_TIME: ClassVar[int] = 18000      # 18000 seconds == 5 hours
    HEADERS: ClassVar[dict[str, str]] = {
        'Content-Type': 'application/x-www-form-urlencoded'
    }
    CHUNK_SIZE: ClassVar[int] = 7000         # Instruments per request

    __slots__ = (
        'public',
        '_private',
        'login',
        '__password',
        '_session_id',
        '_session_time'
    )

    def __init__(
        self,
        public: str | None = None,
        private: str | None = None,
        login: str | None = None,
        password: str | None = None
    ) -> None:
        super().__init__()
        self.logger = getLogger(self.__class__.__name__)

        # Setting authorization data
        self.public = public
        self._private = private
        self.login = login
        self._password = password

        # Checking input
        if not self.public or not self._private:
            self.logger.warning(
                'A keypair was not set. It can be generated here: '
                'https://tradernet.ru/tradernet-api/auth-api'
            )
        if not self.login or not self._password:
            self.logger.warning(
                'Login or password not set. '
                'Some features may be not available.'
            )

        self._session_id: str | None = None
        self._session_time: datetime | None = None

    @classmethod
    def from_config(cls: Type[Self], config_file: str) -> Self:
        """
        Getting a session ID with the use of the login-password
        authorization.

        Parameters
        ----------
        config_file : str
            A path to the configuration file.

        Returns
        -------
        Self
            A new instance.
        """
        config = ConfigParser()
        config.read(config_file)

        return cls(
            config['auth']['public'],
            config['auth']['private'],
            config['auth']['login'],
            config['auth']['password']
        )

    @classmethod
    def from_instance(cls: Type[Self], instance: TraderNetCore) -> Self:
        """
        Creating a new instance from another one.

        Parameters
        ----------
        instance : TraderNetCore
            Other instance to initialize from.

        Returns
        -------
        Self
            A new instance.
        """
        # pylint: disable=protected-access
        core = cls(
            instance.public,
            instance._private,
            instance.login,
            instance._password
        )
        # Avoiding out of sync sessions
        if not instance._session_id:
            instance.get_authorized()
        core._session_id = instance._session_id
        core._session_time = instance._session_time

        return core

    @property
    def url(self) -> str:
        return f'https://{self.DOMAIN}'

    @property
    def websocket_url(self) -> str:
        return f'wss://wss.{self.DOMAIN}'

    def get_websocket_url(self) -> str:
        if not self._session_id:
            if self.login and self._password:
                # Trying to get a session key
                self.get_authorized()
                return self.get_websocket_url()
            # In the case of no success using demo access
            return self.websocket_url
        return f'{self.websocket_url}?SID={self._session_id}'

    def get_auth_info(self) -> dict[str, Any]:
        """
        Getting information about an opened session.

        Returns
        -------
        dict[str, Any]
            Information about the authorization.
        """
        cmd = 'getSidInfo'
        return self.authorized_request(cmd, version=2)

    def get_authorized(self) -> None:
        """
        Getting a session ID with the use of the login-password authorization.
        """
        auth_info = self.get_auth_info()
        # Trying to reuse a session
        if 'SID' in auth_info and auth_info['SID'] and self._session_time and (
            datetime.now() - self._session_time
        ).total_seconds() < self.SESSION_TIME:  # is the session expired?
            self._session_id = auth_info['SID']
            self.logger.debug('Session ID: %s', self._session_id)
            return

        url = f'{self.url}/api/check-login-password'
        message = {
            'login': self.login, 'password': self._password, 'remember_me': 1
        }
        response = self.request('post', url, params=message)
        result = response.json()
        self.logger.debug('Authorization result: %s', result)
        # Setting session key
        self._session_id = result['SID']
        # Setting timer
        self._session_time = datetime.now()

    def plain_request(
        self,
        cmd: str,
        params: dict[str, Any] | None = None
    ) -> Any:
        """
        Unencoded GET request to TraderNet. It could use either use
        authorization or not (if the session ID is not set).

        Parameters
        ----------
        cmd : str
            A command.
        params : dict[str, Any] | None, optional
            Set of parameters in the request.

        Returns
        -------
        Any
            JSON-decoded answer from TraderNet.
        """
        self.logger.debug('Making a simple request to API')

        message = self.__compose_message(cmd, params)
        if self._session_id:
            message['SID'] = self._session_id
            self.logger.debug('Using authorization')

        url = f'{self.url}/api'
        query = {'q': json_dumps(message)}

        self.logger.debug('Message: %s', message)
        self.logger.debug('Query: %s', query)

        response = self.request('get', url, params=query)
        return response.json()

    def authorized_request(
        self,
        cmd: str,
        params: dict[str, Any] | None = None,
        version: int | None = 2
    ) -> Any:
        """
        Sending formatted and encoded request to TraderNet using keypair
        authorization.

        Parameters
        ----------
        cmd : str
            A command.
        params : dict, optional
            Set of parameters in the request.
        version : int, optional
            API version, by default 2

        Returns
        -------
        Answer from TraderNet.
        """
        self.logger.debug('Making request to v%s API with auth', version)
        assert self._private is not None, 'Private key is not set'

        message = self.__compose_message(cmd, params)
        message['nonce'] = int(datetime.now().timestamp() * 10000)

        url = f'{self.url}/api'
        headers = None

        if not version:
            if not self._session_id:
                self.get_authorized()
            message['SID'] = self._session_id
            query: bytes | dict[str, str] = {'q': json_dumps(message)}

        elif version == 1:
            message['sig'] = self.sign(self._private)
            query = {'q': json_dumps(message)}

        elif version == 2:
            url = f'{self.url}/api/v2/cmd/{cmd}'
            message['apiKey'] = self.public
            # Signing the body of the request
            message_string = self.str_from_dict(message)

            headers = deepcopy(self.HEADERS)
            headers['X-NtApi-Sig'] = self.sign(self._private, message_string)
            query = self.http_build_query(message).encode('utf-8')
            self.logger.debug('Message string: %s', message_string)

        else:
            raise ValueError('Unknown API version')

        # Making proper lists of parameters
        if params:
            for key, value in params.items():
                if isinstance(value, list):
                    params[key] = '+'.join(value)
        self.logger.debug(
            'Sending POST to %s, parameters: %s, query: %s', url, params, query
        )
        response = self.request(
            'post', url, headers=headers, params=params, data=query
        )
        return response.json()

    def list_security_sessions(self) -> dict[str, Any]:
        """
        Getting a list of open security sessions of clients.

        Notes
        -----
        https://tradernet.ru/tradernet-api/security-get-list
        """
        cmd = 'getSecuritySessions'
        if not self._session_id:
            self.get_authorized()

        return self.plain_request(cmd)

    def send_security_sms(
        self,
        telegram_only: bool = False,
        push_only: bool = False
    ) -> dict[str, Any]:
        """
        Requesting a security code via SMS.

        Parameters
        ----------
        telegram_only : bool | None
            Send Telegram bot notifications only.
        push_only : bool | None
            Send Push notifications only (for mobile devices).

        Notes
        -----
        https://tradernet.ru/tradernet-api/security-open-sms
        """
        assert not (telegram_only and push_only), \
            'Please choose only one option'

        cmd = 'getSecuritySms'
        params = {'telegram_only': telegram_only, 'push_only': push_only}

        return self.authorized_request(cmd, params, version=None)

    def open_with_sms(
        self,
        token: str
    ) -> dict[str, Any]:
        """
        Opening a security session with a security code sent via SMS.

        Parameters
        ----------
        token : str
            A security code.
        """
        cmd = 'openSecuritySession'
        params = {'validationKey': token, 'safetyTypeId': 3}
        if not self._session_id:
            self.get_authorized()

        return self.authorized_request(cmd, params, version=None)

    def open_with_token(
        self,
        token: str,
        digital_signature: bool = False
    ) -> dict[str, Any]:
        """
        Opening a session with a web token or a digital signature.

        Parameters
        ----------
        token : str
            A signature token.
        digital_signature : bool
            A flag indicating whether the token is a digital signature or a web
            token.

        Notes
        -----
        https://tradernet.ru/tradernet-api/security-open-web-token
        https://tradernet.ru/tradernet-api/security-open-eds
        """
        cmd = 'openSecuritySession'
        if not self._session_id:
            self.get_authorized()

        params = {
            'safetyTypeId': 7 if digital_signature else 8,
            'signature': token,
            'message': self._session_id
        }
        return self.authorized_request(cmd, params, version=None)

    @staticmethod
    def __compose_message(
        cmd: str,
        params: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        message: dict[str, Any] = {'cmd': cmd}
        if params:
            message['params'] = params

        return message



"""
TraderNet market data
"""
from __future__ import annotations

from datetime import datetime
from typing import Any

from pandas import DataFrame, to_datetime

from tradernet.client import TraderNetAPI
from tradernet.symbols.base_market_symbol import BaseMarketSymbol


class TraderNetSymbol(BaseMarketSymbol):
    """
    Acquiring and processing data from TraderNet.

    Parameters
    ----------
    symbol : str
        A symbol name on a remote service.
    start : datetime
        The first date of the period market data to be acquired within.
    end : datetime
        The last date of the period.

    Attributes
    ----------
    symbol : str
        A symbol name on a remote service.
    start : datetime
        The first date of the period market data to be acquired within.
    end : datetime
        The last date of the period.
    market_data : pandas.DataFrame
        A dataframe containing timeseries of candles: open, close, high, low.
    timeframe : int
        Timeframe of candles in seconds. Default is 86400 corresponding to day
        candles.
    """
    __slots__ = ('timeframe',)

    def __init__(
        self,
        symbol: str,
        api: TraderNetAPI | None = None,
        start: datetime = datetime(1970, 1, 1),
        end: datetime = datetime.utcnow()
    ) -> None:
        super().__init__(symbol, api, start, end)
        self.timeframe = 86400

    def __parse_candles(self, candles: dict[str, Any]) -> DataFrame:
        timestamps = \
            [*map(datetime.fromtimestamp, candles['xSeries'][self.symbol])]
        volumes = candles['vl'][self.symbol]
        data = DataFrame(
            candles['hloc'][self.symbol],
            columns=['high', 'low', 'open', 'close']
        )
        data = data.assign(volume=volumes)
        data = data.assign(date=to_datetime(timestamps).round('D'))
        return data

    def get_data(self) -> None:
        if not self.api or not isinstance(self.api, TraderNetAPI):
            self.api = TraderNetAPI()

        candles = self.api.get_candles(
            self.symbol, timeframe=self.timeframe,
            start=self.start, end=self.end
        )

        if 'hloc' in candles:
            data = self.__parse_candles(candles)
        else:
            return
        self.market_data = data.set_index('date').sort_index()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tradernet', 'tradernet.symbols']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0',
 'certifi>=2022.06.15,<2023.0.0',
 'lxml>=4.9.1,<5.0.0',
 'mypy-extensions>=0.4.3,<0.5.0',
 'numpy>=1.23.0,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'tabulate>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'tradernet-sdk',
    'version': '0.3.0',
    'description': 'Public API for TraderNet',
    'long_description': "# tradernet-sdk\n\nPublic Python API for TraderNet\n\n## Installation\n\nInstalling tradernet with pip is straightforward:  \n`python -m pip install tradernet-sdk`  \nInstead of `python` you can use here and further `pypy3` depending on your preferences.\n\n## Usage\n\nImport the client library into your script:  \n`from tradernet import TraderNetAPI`  \nInitialize it with your credentials:  \n`api = TraderNetAPI('public_key', 'private_key', 'login', 'passwd')`  \nor create a config file `tradernet.ini` with the following content:  \n```\n[auth]\npublic   = public_key\nprivate  = private_key\nlogin    = login\npassword = passwd\n```\nand initialize the client with `api = TraderNetAPI.from_config('tradernet.ini')`  \nCall any of its public methods, for example:  \n`api.user_info()`  \n\n### How to trade\n\nUsage of the trading interface is similar to common API. Import and instantiate `Trading` class:  \n```\nfrom tradernet import Trading\n\n\norder = Trading.from_config('tradernet.ini')\n```\nNow let's buy 1 share of FRHC.US at the market price:  \n```\norder.buy('FRHC.US')\n```\n\n### Advanced techniques\n\nOne can import the core class to write their own methods:\n```\nfrom tradernet import TraderNetCore\n\n\nclass MyTNAPI(TraderNetCore):\n    pass\n```\nThis allows using sophisticated request methods for TN API like\n`TraderNetCore.authorized_request`.  \n\nOne can have several instances of the API serving different purposes:  \n```\nconfig = TraderNetCore.from_config('tradernet.ini')\norder = Trading.from_instance(config)\n```\nThe instance `config` stores the credentials, and `order` can be used to trade and may be destroyed after trades completed while `config` is still can be used to instantiate other classes.\n\n\n### Websockets\n\nWebsocket API can be accessed via another class `TraderNetWSAPI`. It\nimplements asynchronous interface for TraderNet API, and its usage is a bit\nmore complicated:  \n```\nfrom tradernet import TraderNetCore, TraderNetWSAPI\n\n\nasync def main() -> None:\n    api = TraderNetCore.from_config('tradernet.ini')\n    async with TraderNetWSAPI(api) as wsapi:  # type: TraderNetWSAPI\n        async for quote in wsapi.market_depth('FRHC.US'):\n            print(quote)\n```\n\n### Legacy API\n\nThe library also has the legacy `PublicApiClient.py` which provides almost\nthe same functionality as most of TraderNet users used to:\n```\nfrom tradernet import NtApi\n\n\npub_ = '[public Api key]'\nsec_ = '[secret Api key]'\ncmd_ = 'getPositionJson'\nres = NtApi(pub_, sec_, NtApi.V2)\nprint(res.sendRequest(cmd_))\n```\nThe only difference is that one doesn't have to decode the content of the\nresponse as before.\n\n### Options\n\nThe notation of options in TraderNet now can easily be deciphered:\n```\nfrom tradernet import TraderNetOption\n\n\noption = TraderNetOption('+FRHC.16SEP2022.C55')\nprint(option)  # FRHC.US @ 55 Call 2022-09-16\n```\nor the scary old notation:\n```\nfrom tradernet import DasOption\n\n\noption = DasOption('+FRHC^C7F45.US')\nprint(option)  # FRHC.US @ 45 Call 2022-07-15\n```\n\n### Wrapping market data\n\nAnother feature is to get handy pandas.DataFrame objects with market data:\n```\nfrom tradernet import TraderNetSymbol, TraderNetAPI\n\n\napi = TraderNetAPI('public_key', 'private_key', 'login', 'passwd')\nsymbol = TraderNetSymbol('AAPL.US', api)\nsymbol.get_data()\nprint(symbol.market_data.head().to_markdown())\n# | date                |     high |      low |     open |    close |      volume |\n# |:--------------------|---------:|---------:|---------:|---------:|------------:|\n# | 1980-12-12 00:00:00 | 0.128876 | 0.12834  | 0.12834  | 0.12834  | 1.17258e+08 |\n# | 1980-12-15 00:00:00 | 0.122224 | 0.121644 | 0.122224 | 0.121644 | 4.39712e+07 |\n# | 1980-12-16 00:00:00 | 0.113252 | 0.112716 | 0.113252 | 0.112716 | 2.6432e+07  |\n# | 1980-12-17 00:00:00 | 0.116064 | 0.115484 | 0.115484 | 0.115484 | 2.16104e+07 |\n# | 1980-12-18 00:00:00 | 0.119412 | 0.118876 | 0.118876 | 0.118876 | 1.83624e+07 |\n```\n\n## License\n\nThe package is licensed under permissive MIT License. See the `LICENSE` file in\nthe top directory for the full license text.\n",
    'author': 'Anton Kudelin',
    'author_email': 'a.kudelin@freedomfinance.eu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.nettrader.ru/risk/python-tradernet-sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from dataclasses import dataclass
from datetime import date
from decimal import Decimal


@dataclass
class OptionProperties:
    ticker: str
    location: str | None
    right: int
    strike: Decimal
    maturity_date: date
    symbolic_expiration: str
